@@ -6,29 +6,46 @@
 
 #include "rimeengine.h"
 #include "notifications_public.h"
+#include "rimeaction.h"
 #include "rimestate.h"
 #include <cstdint>
 #include <cstring>
 #include <ctime>
 #include <dirent.h>
+#include <exception>
+#include <fcitx-config/iniparser.h>
+#include <fcitx-config/rawconfig.h>
 #include <fcitx-utils/event.h>
 #include <fcitx-utils/fs.h>
 #include <fcitx-utils/i18n.h>
 #include <fcitx-utils/log.h>
+#include <fcitx-utils/macros.h>
 #include <fcitx-utils/misc.h>
 #include <fcitx-utils/standardpath.h>
 #include <fcitx-utils/stringutils.h>
+#include <fcitx/action.h>
+#include <fcitx/addoninstance.h>
 #include <fcitx/candidatelist.h>
+#include <fcitx/event.h>
 #include <fcitx/inputcontext.h>
 #include <fcitx/inputcontextmanager.h>
+#include <fcitx/inputmethodentry.h>
 #include <fcitx/inputpanel.h>
+#include <fcitx/instance.h>
+#include <fcitx/menu.h>
 #include <fcitx/statusarea.h>
 #include <fcitx/userinterface.h>
 #include <fcitx/userinterfacemanager.h>
-#include <optional>
+#include <list>
+#include <memory>
 #include <rime_api.h>
 #include <stdexcept>
 #include <string>
+#include <string_view>
+#include <thread>
+#include <unordered_map>
+#include <utility>
+#include <vector>
 
 FCITX_DEFINE_LOG_CATEGORY(rime, "rime");
 
@@ -152,144 +169,6 @@ private:
     RimeEngine *engine_;
 };
 
-class ToggleAction : public RimeOptionAction {
-public:
-    ToggleAction(RimeEngine *engine, std::string_view schema,
-                 std::string_view option, std::string disabledText,
-                 std::string enabledText)
-        : engine_(engine), option_(option),
-          disabledText_(std::move(disabledText)),
-          enabledText_(std::move(enabledText)) {
-        engine_->instance()->userInterfaceManager().registerAction(
-            stringutils::concat("fcitx-rime-", schema, "-", option), this);
-    }
-
-    void activate(InputContext *ic) override {
-        auto *state = engine_->state(ic);
-        auto *api = engine_->api();
-        if (!state) {
-            return;
-        }
-        // Do not send notification since user is explicitly select it.
-        engine_->blockNotificationFor(30000);
-        auto session = state->session();
-        Bool oldValue = api->get_option(session, option_.c_str());
-        api->set_option(session, option_.c_str(), !oldValue);
-    }
-
-    std::string shortText(InputContext *ic) const override {
-        auto *state = engine_->state(ic);
-        auto *api = engine_->api();
-        if (!state) {
-            return "";
-        }
-        auto session = state->session();
-        if (api->get_option(session, option_.c_str())) {
-            return stringutils::concat(enabledText_, " → ", disabledText_);
-        }
-        return stringutils::concat(disabledText_, " → ", enabledText_);
-    }
-
-    std::string icon(InputContext * /*unused*/) const override { return ""; }
-
-    std::optional<std::string> snapshotOption(InputContext *ic) override {
-        auto *state = engine_->state(ic);
-        auto *api = engine_->api();
-        if (!state) {
-            return std::nullopt;
-        }
-        auto session = state->session(false);
-        if (!session) {
-            return std::nullopt;
-        }
-        if (!api->get_option(session, option_.c_str())) {
-            return stringutils::concat("!", option_);
-        }
-        return option_;
-    }
-
-private:
-    RimeEngine *engine_;
-    std::string option_;
-    std::string disabledText_;
-    std::string enabledText_;
-};
-
-class SelectAction : public RimeOptionAction {
-public:
-    SelectAction(RimeEngine *engine, std::string_view schema,
-                 std::vector<std::string> options,
-                 std::vector<std::string> texts)
-        : engine_(engine), options_(options), texts_(std::move(texts)) {
-        for (size_t i = 0; i < options.size(); ++i) {
-            actions_.emplace_back();
-            actions_.back().setShortText(texts_[i]);
-            actions_.back().connect<SimpleAction::Activated>(
-                [this, i](InputContext *ic) {
-                    auto *state = engine_->state(ic);
-                    auto *api = engine_->api();
-                    if (!state) {
-                        return;
-                    }
-                    auto session = state->session();
-                    for (size_t j = 0; j < options_.size(); ++j) {
-                        api->set_option(session, options_[j].c_str(), i == j);
-                    }
-                });
-            engine_->instance()->userInterfaceManager().registerAction(
-                stringutils::concat("fcitx-rime-", schema, "-", options_[i]),
-                &actions_.back());
-            menu_.addAction(&actions_.back());
-        }
-        setMenu(&menu_);
-        engine_->instance()->userInterfaceManager().registerAction(
-            stringutils::concat("fcitx-rime-", schema, "-select-", options[0]),
-            this);
-    }
-
-    std::string shortText(InputContext *ic) const override {
-        auto *state = engine_->state(ic);
-        auto *api = engine_->api();
-        if (!state) {
-            return "";
-        }
-        auto session = state->session();
-        for (size_t i = 0; i < options_.size(); ++i) {
-            if (api->get_option(session, options_[i].c_str())) {
-                return texts_[i];
-            }
-        }
-        return "";
-    }
-
-    std::string icon(InputContext * /*unused*/) const override { return ""; }
-
-    std::optional<std::string> snapshotOption(InputContext *ic) override {
-        auto *state = engine_->state(ic);
-        auto *api = engine_->api();
-        if (!state) {
-            return std::nullopt;
-        }
-        auto session = state->session(false);
-        if (!session) {
-            return std::nullopt;
-        }
-        for (size_t i = 0; i < options_.size(); ++i) {
-            if (api->get_option(session, options_[i].c_str())) {
-                return options_[i];
-            }
-        }
-        return std::nullopt;
-    }
-
-private:
-    RimeEngine *engine_;
-    std::vector<std::string> options_;
-    std::vector<std::string> texts_;
-    std::list<SimpleAction> actions_;
-    Menu menu_;
-};
-
 bool RimeEngine::firstRun_ = true;
 
 RimeEngine::RimeEngine(Instance *instance)
@@ -524,6 +403,9 @@ void RimeEngine::updateConfig() {
     updateSchemaMenu();
     refreshSessionPoolPolicy();
 
+    deployAction_.setHotkey(config_.deploy.value());
+    syncAction_.setHotkey(config_.synchronize.value());
+
     if (constructed_) {
         refreshStatusArea(0);
     }
@@ -593,15 +475,29 @@ void RimeEngine::activate(const InputMethodEntry & /*entry*/,
                           InputContextEvent &event) {
     auto *ic = event.inputContext();
     refreshStatusArea(*ic);
+    if (auto *state = this->state(ic)) {
+        state->activate();
+    }
 }
 
 void RimeEngine::deactivate(const InputMethodEntry &entry,
                             InputContextEvent &event) {
-    if (event.type() == EventType::InputContextSwitchInputMethod &&
-        *config_.commitWhenDeactivate) {
+    if (event.type() == EventType::InputContextSwitchInputMethod) {
         auto *inputContext = event.inputContext();
         auto *state = this->state(inputContext);
-        state->commitPreedit(inputContext);
+        switch (*config_.switchInputMethodBehavior) {
+        case SwitchInputMethodBehavior::Clear:
+            break;
+        case SwitchInputMethodBehavior::CommitRawInput:
+            state->commitInput(inputContext);
+            break;
+        case SwitchInputMethodBehavior::CommitComposingText:
+            state->commitComposing(inputContext);
+            break;
+        case SwitchInputMethodBehavior::CommitCommitPreview:
+            state->commitPreedit(inputContext);
+            break;
+        }
     }
     reset(entry, event);
 }
@@ -612,8 +508,19 @@ void RimeEngine::keyEvent(const InputMethodEntry &entry, KeyEvent &event) {
     RIME_DEBUG() << "Rime receive key: " << event.rawKey() << " "
                  << event.isRelease();
     auto *inputContext = event.inputContext();
+    if (!event.isRelease()) {
+        if (event.key().checkKeyList(*config_.deploy)) {
+            deploy();
+            return event.filterAndAccept();
+        } else if (event.key().checkKeyList(*config_.synchronize)) {
+            sync();
+            return event.filterAndAccept();
+        }
+    }
     auto *state = this->state(inputContext);
+    currentKeyEventState_ = state;
     state->keyEvent(event);
+    currentKeyEventState_ = nullptr;
 }
 
 void RimeEngine::reset(const InputMethodEntry & /*entry*/,
@@ -642,7 +549,10 @@ void RimeEngine::rimeNotificationHandler(void *context, RimeSessionId session,
                                          const char *messageValue) {
     RIME_DEBUG() << "Notification: " << session << " " << messageType << " "
                  << messageValue;
-    RimeEngine *that = static_cast<RimeEngine *>(context);
+    auto *that = static_cast<RimeEngine *>(context);
+    if (that->mainThreadId_ == std::this_thread::get_id()) {
+        that->notifyImmediately(session, messageType, messageValue);
+    }
     that->eventDispatcher_.schedule(
         [that, session, messageType = std::string(messageType),
          messageValue = std::string(messageValue)]() {
@@ -650,12 +560,26 @@ void RimeEngine::rimeNotificationHandler(void *context, RimeSessionId session,
         });
 }
 
+void RimeEngine::notifyImmediately(RimeSessionId session,
+                                   std::string_view messageType,
+                                   std::string_view messageValue) {
+    if (messageType != "option") {
+        return;
+    }
+    if (!currentKeyEventState_ ||
+        currentKeyEventState_->session(false) != session) {
+        return;
+    }
+    currentKeyEventState_->addChangedOption(messageValue);
+}
+
 void RimeEngine::notify(RimeSessionId session, const std::string &messageType,
                         const std::string &messageValue) {
     const char *message = nullptr;
     const char *icon = "";
     const char *tipId = "";
     int timeout = 3000;
+    bool blockMessage = false;
     if (messageType == "deploy") {
         tipId = "fcitx-rime-deploy";
         icon = "fcitx_rime_deploy";
@@ -673,36 +597,14 @@ void RimeEngine::notify(RimeSessionId session, const std::string &messageType,
             }
             updateSchemaMenu();
             refreshStatusArea(0);
+            blockMessage = true;
         } else if (messageValue == "failure") {
             needRefreshAppOption_ = false;
             message = _("Rime has encountered an error. "
                         "See log for details.");
+            blockMessage = true;
         }
     } else if (messageType == "option") {
-        // Only show option notification triggered by key event.
-        if (!isAndroid() && lastKeyEventTime_ + 30000 > now(CLOCK_MONOTONIC)) {
-            icon = "fcitx-rime";
-            if (messageValue == "!full_shape") {
-                tipId = "fcitx-rime-full-shape";
-                message = _("Half Shape is enabled.");
-            } else if (messageValue == "full_shape") {
-                tipId = "fcitx-rime-full-shape";
-                message = _("Full Shape is enabled.");
-            } else if (messageValue == "!ascii_punct") {
-                tipId = "fcitx-rime-ascii-punct";
-                message = _("Punctuation conversion is enabled.");
-            } else if (messageValue == "ascii_punct") {
-                tipId = "fcitx-rime-ascii-punct";
-                message = _("Punctuation conversion is disabled.");
-            } else if (messageValue == "!simplification") {
-                tipId = "fcitx-rime-simplification";
-                message = _("Traditional Chinese is enabled.");
-            } else if (messageValue == "simplification") {
-                tipId = "fcitx-rime-simplification";
-                message = _("Simplified Chinese is enabled.");
-            }
-        }
-        timeout = 1000;
         updateStatusArea(session);
     } else if (messageType == "schema") {
         // Schema is changed either via status area or shortcut
@@ -715,6 +617,10 @@ void RimeEngine::notify(RimeSessionId session, const std::string &messageType,
         notifications->call<INotifications::showTip>(
             tipId, _("Rime"), icon, _("Rime"), message, timeout);
     }
+    // Block message after error / success.
+    if (blockMessage) {
+        blockNotificationFor(30000);
+    }
 }
 
 RimeState *RimeEngine::state(InputContext *ic) {
